from utils import *
import config as cfg
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk

client = Elasticsearch()
index_name = 'testing'
data_dir = '/path/to/images'

# Initialize elasticsearch classes
idx = Index(index_name, cfg.index_file, cfg.data, client)
doc = Document(data_dir, cfg.data, index_name)

# Create an index
print('Creating index ', index_name)
idx.createIndex()

print('Creating documents..')
doc.run()

print('Bulk indexing..')
idx.bulkIndex()

print('Done!')



import sys
sys.path.append('webapp/')
from utils import *
import config as cfg

app = True

if app:
    from photoscope.webapp.app import app
    app.secret_key = 'super secret key'
    app.run("0.0.0.0", port=5000)



import os
import json
import argparse
import numpy as np
from photoscope.labels import labels

from PIL import Image
import tensorflow_hub as hub
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk


class Index(object):
    def __init__(self, index_name, index_file, data, client):
        self.index_name = index_name
        self.index_file = index_file
        self.data = data

        self.client = client

    def loadDataset(self, path):
        with open(path) as f:
            return [json.loads(line) for line in f]

    def createIndex(self):
        self.client.indices.delete(index=self.index_name, ignore=[404])
        with open(self.index_file) as index_file:
            source = index_file.read().strip()
            self.client.indices.create(index=self.index_name, body=source)

    def bulkIndex(self):
        docs = self.loadDataset(self.data)
        bulk(self.client, docs)

    def index(self, doc):
        self.client.index(index=self.index_name, body=doc)


class Document(object):
    def __init__(self, image_dir, data, index_name):
        self.image_dir = image_dir
        self.data = data
        self.index_name = index_name

        self.feature_extractor_url = "https://tfhub.dev/google/imagenet/mobilenet_v1_050_160/feature_vector/4" 
        self.classifier_url = "https://tfhub.dev/google/imagenet/mobilenet_v1_050_160/classification/4" 
        self.sentence_encoder_url = "https://tfhub.dev/google/universal-sentence-encoder/4"

        self.feature_extractor_layer = hub.KerasLayer(self.feature_extractor_url, input_shape=(224,224,3))
        self.classifier = hub.KerasLayer(self.classifier_url, input_shape=(224,224,3))
        self.sentence_encoder = hub.load(self.sentence_encoder_url)


    def createDocument(self, doc, emb, tag, tag_vector, index_name):
        return {
            '_op_type': 'index',
            '_index': index_name,
            'filepath': doc['image'],
            'image_vector': emb,
            'tag': tag,
            'tag_vector': tag_vector,
        }


    def loadDataset(self, path):
        docs = []
        files = os.listdir(path)
        for fl in files:
            doc = {
                'image': path + fl,
            }
            docs.append(doc)
        return docs
    

    def bulkPredict(self, docs, batch_size=32):
        """Extract mobilenet embeddings."""
        for i in range(0, len(docs), batch_size):
            batch_docs = docs[i: i+batch_size]
            batch_docs = [Image.open(doc['image']) for doc in batch_docs]
            batch_docs = np.array([(np.array(doc.resize((224,224)))/255).astype(np.float32) for doc in batch_docs])

            embeddings = self.feature_extractor_layer(batch_docs)
            tags = self.classifier(batch_docs)

            tag_sentences = [x.argsort()[-5:][::-1] for x in tags.numpy()] 
            tag_sentences = [', '.join([labels[i] for i in sublist]) for sublist in tag_sentences]

            tag_embeddings = self.sentence_encoder(tag_sentences)

            for emb, tag, tag_emb in zip(embeddings, tag_sentences, tag_embeddings):
                yield emb.numpy().tolist(), tag, tag_emb.numpy().tolist()


    def run(self):
        docs = self.loadDataset(self.image_dir)
        with open(self.data, 'w') as f:
            for doc, preds in zip(docs, self.bulkPredict(docs)):
                d = self.createDocument(doc, preds[0], preds[1], preds[2], self.index_name)
                f.write(json.dumps(d) + '\n')
                f.flush()



from setuptools import setup, find_packages 
  
long_description = 'Simple command line tool for launching an image search app using elasticsearch and mobilenet embeddings.' 
  
setup( 
        name ='photoscope', 
        version ='1.0.1', 
        author ='Smells Like ML', 
        author_email ='contact@smellslikeml.com', 
        url ='https://github.com/smellslikeml/photoscope', 
        description ='Elasticsearch + Flask for diy image search', 
        long_description = long_description, 
        long_description_content_type ="text/markdown", 
        license ='MIT', 
        packages = find_packages(), 
        entry_points ={ 
            'console_scripts': [ 
                'photoscope = photoscope.main:main'
            ] 
        }, 
        classifiers =( 
            "Programming Language :: Python :: 3", 
            "License :: OSI Approved :: MIT License", 
            "Operating System :: OS Independent", 
        ), 
        keywords ='smellslikeml, machine learning, search', 
        include_package_data=True,
        install_requires = [
            'tensorflow==2.1.0',
            'tensorflow_hub==0.7.0',
            'elasticsearch==7.6.0',
            'pillow==7.0.0',
            'flask==1.1.1',
            'requests'
            ], 
        zip_safe = False
) 




import os
import time
import json
import base64
import requests
import numpy as np
from glob import glob
from PIL import Image
from io import BytesIO
from flask import Flask
from flask import render_template, request, send_from_directory, jsonify, redirect, url_for, session
from elasticsearch import Elasticsearch
import photoscope.config as cfg
from photoscope.labels import labels
from photoscope.utils import Index, Document
from werkzeug.utils import secure_filename


app = Flask(__name__)

client = Elasticsearch()

temp_dir= 'tmp/'

indexer = Index(cfg.index_name, cfg.index_file, cfg.doc_json, client)
doc = Document(cfg.data_dir, cfg.doc_json, cfg.index_name)

image_encoder = doc.feature_extractor_layer
classifier = doc.classifier
sentence_encoder = doc.sentence_encoder

@app.route('/', methods=['GET', 'POST'])
def home():
    if request.method == "POST":
        data_file = request.files.getlist('file[]')[0]
        url = request.form.get('url')
        tags = request.form.get('tag')
        if data_file:
            file_contents = data_file.read()
        elif url:
            response = requests.get(url)
            file_contents = response.content

        if data_file or url:
            image = Image.open(BytesIO(file_contents))
            image = (np.array(image.resize((224,224)))/255).astype(np.float32)
            img_array = np.expand_dims(image, axis=0)

            embedding = image_encoder(img_array)
            query_vector = embedding.numpy().tolist()[0]
            script_query = {
                "script_score": {
                    "query": {"match_all": {}},
                    "script": {
                        "source": "cosineSimilarity(params.query_vector, doc['image_vector']) + 1.0",
                        "params": {"query_vector": query_vector}
                    }
                }
            }
        elif tags:
            sentence_embedding = sentence_encoder([tags])
            query_vector = sentence_embedding.numpy().tolist()[0]
            script_query = {
                "script_score": {
                    "query": {"match_all": {}},
                    "script": {
                        "source": "cosineSimilarity(params.query_vector, doc['tag_vector']) + 1.0",
                        "params": {"query_vector": query_vector}
                    }
                }
            }


        response = client.search(
            index=cfg.index_name,
            body={
                "size": cfg.search_size,
                "query": script_query,
                "_source": {"includes": ["filepath", "tags"]}
            }
        )
        result_paths = [x['_source']['filepath'] for x in response['hits']['hits']]
        result_paths = [pth.split('/')[-1] for pth in result_paths]
        return render_template("results.html", result_paths=result_paths)
    return render_template("index.html")

def allowed_file(filename):
    return '.' in filename and \
           filename.rsplit('.', 1)[1].lower() in ALLOWED_EXTENSIONS


@app.route('/upload', methods=['GET', 'POST'])
def upload():
    if request.method == "POST":
        data_files = request.files.getlist('file[]')
        images_list = []
        tags_list = []
        for pic in data_files:
            file_contents = pic.read()

            # Generate tags for image
            img = Image.open(BytesIO(file_contents))
            img.save(temp_dir + pic.filename)
            image = (np.array(img.resize((224,224)))/255).astype(np.float32)
            img_array = np.expand_dims(image, axis=0)

            embedding = classifier(img_array)
            tags = embedding.numpy()[0].argsort()[-5:][::-1] 
            tags = ','.join([labels[i] for i in tags])

            images_list.append(pic.filename)
            tags_list.append(tags)
        session["images_list"] = images_list
        session["tags_list"] = tags_list
        return redirect(url_for("confirm"))
    return render_template("upload.html")


@app.route('/confirm', methods=['GET', 'POST'])
def confirm():
    images_list = session.get("images_list")
    tags_list = session.get("tags_list")
    if request.method == "POST":
        for i in range(len(images_list)):
            tags = request.form.get('tokenfield{}'.format(i)) #getting all confirmed tags
            sent_embedding = sentence_encoder([tags])

            img = Image.open(temp_dir + images_list[i])
            img = (np.array(img.resize((224,224)))/255).astype(np.float32)
            img_array = np.expand_dims(img, axis=0)
            img_embedding = image_encoder(img_array)

            image_vector = img_embedding.numpy().tolist()[0]
            tag_vector = sent_embedding.numpy().tolist()[0]

            new_doc = {'filepath': str(os.path.join(cfg.data_dir, images_list[i])),
                    'image_vector': image_vector,
                    'tag': str(tags),
                    'tag_vector': tag_vector}

            os.rename(os.path.join(temp_dir, images_list[i]), os.path.join(cfg.data_dir, images_list[i]))

            indexer.index(json.dumps(new_doc))
            return redirect(url_for("home"))
    return render_template("confirm.html", images_list=images_list, tags_list=tags_list)

@app.route('/t/<path:filename>')
def get_temp(filename):
    return send_from_directory(temp_dir, filename)

@app.route('/image/<path:filename>')
def get_image(filename):
    return send_from_directory(cfg.data_dir, filename)

if __name__ == "__main__":
    app.secret_key = 'super secret key'
    app.run("0.0.0.0", port=5000)



import os
import logging
import argparse
import photoscope.config as cfg
from photoscope.utils import Index, Document

CURRENT_DIR = os.path.dirname(__file__)
os.chdir(CURRENT_DIR)


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='subcommand')

    #  subparser for configuration
    parser_configure = subparsers.add_parser('configure')
    # add required arguments
    parser_configure.add_argument('--data_dir', type=str, help='directory of images to index and serve', required=True)
    parser_configure.add_argument('--index_name', type=str, help='index name for your search engine', default=cfg.index_name)
    parser_configure.add_argument('--search_size', type=int, help='number of search results to present', default=cfg.search_size)
    parser_configure.add_argument('--index_file', type=str, help='index json file for index structure', default=cfg.index_file)
    parser_configure.add_argument('--doc_json', type=str, help='document json file for name bulk indexing', default=cfg.doc_json)


    #  subparser for index
    parser_index = subparsers.add_parser('index')
    # add required arguments
    parser_index.add_argument('--index_name', type=str, help='index name for your search engine', default=cfg.index_name)
    parser_index.add_argument('--data_dir', type=str, help='directory of images to index and serve', default=cfg.data_dir)


    #  subparser for app
    parser_index = subparsers.add_parser('app')
    # add required arguments
    parser_index.add_argument('start', help='start the elasticsearch app')


    args = parser.parse_args()
    if args.subcommand == 'configure':
        new_vars = ["data_dir='{}'".format(str(args.data_dir)),
                    "search_size='{}'".format(str(args.search_size)),
                    "index_name='{}'".format(str(args.index_name)),
                    "index_file='{}'".format(str(args.index_file)),
                    "doc_json='{}'".format(str(args.doc_json))]
        with open('config.py', 'w') as f:
            for var in new_vars:
                f.write(var+'\n')

    elif args.subcommand == 'index':
        assert (cfg.data_dir!=None), "The data_dir flag has not been set, please run: photoscope configure --data_dir='/path/to/imgs/'"
        from elasticsearch import Elasticsearch
        from elasticsearch.helpers import bulk

        client = Elasticsearch()

        # Initialize elasticsearch classes
        idx = Index(cfg.index_name, cfg.index_file, cfg.doc_json, client)
        doc = Document(cfg.data_dir, cfg.doc_json, cfg.index_name)

        # Create an index
        logging.info('Creating index {}'.format(str(cfg.index_name)))
        idx.createIndex()

        logging.info('Creating documents...')
        doc.run()

        logging.info('Bulk indexing documents...')
        idx.bulkIndex()

        logging.info('DONE!')
        
    elif args.subcommand == 'app':
        assert (cfg.data_dir!=None), "The data_dir flag has not been set, please run: photoscope configure --data_dir='/path/to/imgs/'"
        os.chdir("webapp/")

        filelist = [ f for f in os.listdir("tmp/")]
        for f in filelist:
            os.remove(os.path.join("tmp/", f))

        from photoscope.webapp.app import app
        app.secret_key = 'super secret key'
        app.run("0.0.0.0", port=5000)
